# -*- coding: utf-8 -*-
#
# Copyright (c) 2017 -- Lars Heuer - Semagia <http://www.semagia.com/>.
# All rights reserved.
#
# License: BSD License
#
"""\
Integration into Flask.
"""
from __future__ import absolute_import, unicode_literals
import segno

# <https://wiki.python.org/moin/PortingToPy3k/BilingualQuickRef#New_Style_Classes>
__metaclass__ = type

